/*===========================================================================*
 * mpeg.c								     *
 *									     *
 *	Procedures to generate the MPEG sequence			     *
 *									     *
 * EXPORTED PROCEDURES:							     *
 *	GetMPEGStream							     *
 *	IncrementTCTime							     *
 *	SetStatFileName							     *
 *	SetGOPSize							     *
 *	PrintStartStats							     *
 *									     *
 *===========================================================================*/

/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/*
 *  modified by Miyake 2001/2/7
 *  $Header: /n/picasso/project/mpeg/mpeg_dist/mpeg_encode/RCS/mpeg.c,v 1.24 1995/08/16 18:10:48 smoot Exp smoot $
 *  $Log: mpeg.c,v $
 *  Revision 1.24  1995/08/16 18:10:48  smoot
 *  *** empty log message ***
 *
 *  Revision 1.23  1995/08/07 21:48:08  smoot
 *  stdin bugs fixed
 *
 *  Revision 1.22  1995/06/26 21:49:19  smoot
 *  added new frame ordering (hacks)^H^H^H^H^H code ;-)
 *
 *  Revision 1.21  1995/06/21 18:30:41  smoot
 *  changed time structure to be ANSI
 *  changed file access to be binary (DOS!)
 *  added time to userdata
 *  Added a sleep to remote reads (NFS delay)
 *
 * Revision 1.20  1995/05/02  01:49:21  eyhung
 * prints out true output bit rate and slightly untabified
 *
 * Revision 1.19  1995/05/02  00:45:35  eyhung
 * endstats now contain correct output fbit rate at the specified frame rate
 *
 * Revision 1.18  1995/03/27  23:43:20  smoot
 * killed printing long as int (compiler warning)
 *
 * Revision 1.17  1995/03/27  19:18:54  smoot
 * fixed divide by zero for very quick encodings
 *
 * Revision 1.16  1995/02/02  22:03:37  smoot
 * added types for MIPS
 *
 * Revision 1.15  1995/02/02  07:26:58  eyhung
 * removed unused tempframe
 *
 * Revision 1.14  1995/02/01  05:01:35  eyhung
 * Completed infinite coding-on-the-fly
 *
 * Revision 1.13  1995/02/01  02:34:02  eyhung
 * Added full coding-on-the-fly
 *
 * Revision 1.12  1995/01/31  23:05:14  eyhung
 * Added some stdin stuff
 *
 * Revision 1.11  1995/01/20  00:01:16  eyhung
 * Added output file to PrintEndStats
 *
 * Revision 1.10  1995/01/19  23:08:51  eyhung
 * Changed copyrights
 *
 * Revision 1.9  1995/01/17  18:55:54  smoot
 * added right version number, and error if no frames selected
 *
 * Revision 1.8  1995/01/16  08:12:54  eyhung
 * added realQuiet
 *
 * Revision 1.7  1994/12/07  00:40:36  smoot
 * Added seperate P and B search ranges
 *
 * Revision 1.6  1994/11/28  21:46:45  smoot
 * Added version printing
 *
 * Revision 1.5  1994/11/19  01:33:05  smoot
 * put in userdata
 *
 * Revision 1.4  1994/11/14  22:36:22  smoot
 * Merged specifics and rate control
 *
 * Revision 1.2  1994/03/15  00:27:11  keving
 * nothing
 *
 * Revision 1.1  1993/12/22  19:19:01  keving
 * nothing
 *
 * Revision 1.6  1993/07/22  22:23:43  keving
 * nothing
 *
 * Revision 1.5  1993/06/30  20:06:09  keving
 * nothing
 *
 * Revision 1.4  1993/06/03  21:08:08  keving
 * nothing
 *
 * Revision 1.3  1993/02/19  18:10:12  keving
 * nothing
 *
 * Revision 1.2  1993/02/17  23:18:20  dwallach
 * checkin prior to keving's joining the project
 *
 */


/*==============*
 * HEADER FILES *
 *==============*/

#include "all.h"
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include "mtypes.h"
#include "frames.h"
#include "search.h"
#include "mpeg.h"
#include "prototypes.h"
#include "parallel.h"
#include "param.h"
#include "readframe.h"
#include "fsize.h"
#include "mheaders.h"
#include "rate.h"
#ifdef MIPS
#include <sys/types.h>
#endif
#include <sys/stat.h>

/*===========*
 *  VERSION  *
 *===========*/

#define VERSION "1.5b"

/*======================*
 *  GtkPBar  (By Miyake)*
 *======================*/
#ifdef GtkPBar
extern boolean update_PBar(float ratio);
#endif


/*===========*
 * CONSTANTS *
 *===========*/

#define	FPS_30	0x5   /* from MPEG standard sect. 2.4.3.2 */
#define ASPECT_1    0x1	/* aspect ratio, from MPEG standard sect. 2.4.3.2 */


/*==================*
 * STATIC VARIABLES *
 *==================*/

static int32   diffTime;
static int framesOutput;
static int	    realStart, realEnd;
static int	currentGOP;
static int	    timeMask;
static int	    numI, numP, numB;


/*==================*
 * GLOBAL VARIABLES *	
 *==================*/

/* important -- don't initialize anything here */
/* must be re-initted anyway in GenMPEGStream */

extern boolean	resizeFrame;
extern int outputWidth, outputHeight;
int	    gopSize = 100;  /* default */
int32	    tc_hrs, tc_min, tc_sec, tc_pict, tc_extra;
int	    totalFramesSent;
int	    yuvWidth, yuvHeight;
int	    realWidth, realHeight;
char	    currentPath[MAXPATHLEN];
char	    statFileName[256];
time_t	    timeStart, timeEnd;
FILE	   *statFile;
FILE	   *bitRateFile = NULL;
char	   *framePattern;
int	    framePatternLen;
int	    referenceFrame;
static int  framesRead;
MpegFrame  *pastRefFrame;
MpegFrame  *futureRefFrame;
int	    frameRate = FPS_30;
int	    frameRateRounded = 30;
boolean	    frameRateInteger = TRUE;
int	    aspectRatio = ASPECT_1;
extern unsigned char userDataFileName[];
extern int mult_seq_headers;

int32 bit_rate, buf_size;

/*===============================*
 * INTERNAL PROCEDURE prototypes *
 *===============================*/

static void	ProcessRefFrame _ANSI_ARGS_((MpegFrame *frame,
					      BitBucket *bb, int lastFrame,
					      char *outputFileName));
static void	OpenBitRateFile _ANSI_ARGS_((void));
static void	CloseBitRateFile _ANSI_ARGS_((void));


/*=====================*
 * EXPORTED PROCEDURES *
 *=====================*/

/*===========================================================================*
 *
 * SetReferenceFrameType
 *
 *	set the reference frame type to be original or decoded
 *
 * RETURNS:	nothing
 *
 * SIDE EFFECTS:    referenceFrame
 *
 *===========================================================================*/
void
SetReferenceFrameType(type)
    char *type;
{
    if ( strcmp(type, "ORIGINAL") == 0 ) 
      referenceFrame = ORIGINAL_FRAME;
    else if ( strcmp(type, "DECODED") == 0 )
      referenceFrame = DECODED_FRAME;
    else 
      exit(1);
}

/*===========================================================================*
 *
 * GenMPEGStream
 *
 *	generate an MPEG sequence stream (generally)
 *	if whichGOP == frameStart == -1 then does complete MPEG sequence
 *	if whichGOP != -1 then does numbered GOP only (without sequence
 *			       header)
 *	if frameStart != -1 then does numbered frames only (without any
 *				 sequence or GOP headers)		       
 *
 * RETURNS:	0 if completed -1 not completed
 *
 * SIDE EFFECTS:    too numerous to mention
 *
 *===========================================================================*/
int32
GenMPEGStream(whichGOP, frameStart, frameEnd, qtable, niqtable, numFrames,
              ofp, outputFileName)
    int whichGOP;
    int frameStart;
    int frameEnd;
    int32   qtable[];
    int32   niqtable[];
    int numFrames;
    FILE *ofp;
    char *outputFileName;
{
    extern void        PrintItoIBitRate _ANSI_ARGS_((int numBits, int frameNum));
    BitBucket *bb;
    int i;
    char frameType;
    MpegFrame        *frame = NULL;
    MpegFrame *tempFrame;
    int            firstFrame, lastFrame;
    int     inputFrameBits = 0;
    char    inputFileName[1024];
    time_t  tempTimeStart, tempTimeEnd;
    boolean firstFrameDone = FALSE;
    long numBits;
    int32 bitstreamMode, res;
    boolean interrupt = FALSE;

    time(&timeStart);

    framesRead = 0;

    ResetIFrameStats();
    ResetPFrameStats();
    ResetBFrameStats();

    Fsize_Reset();

    framesOutput = 0;

    SetFileType(inputConversion);

    firstFrame = 0;
    lastFrame = numFrames-1;

    realStart = 0;
    realEnd = numFrames-1;
    if ( numFrames == 0 )  
      exit(1);

    /* count number of I, P, and B frames */
    numI = 0;        numP = 0;   numB = 0;
    timeMask = 0;
 
    for ( i = firstFrame; i <= lastFrame; i++ ) 
    {
       frameType = FType_Type(i);
       switch(frameType) 
       {
	 case 'i':        numI++;            timeMask |= 0x1;    break;
	 case 'p':        numP++;            timeMask |= 0x2;        break;
	 case 'b':        numB++;            timeMask |= 0x4;        break;
       }
     }

    if ( frameStart == -1 )
      bb = Bitio_New(ofp);

    totalFramesSent = firstFrame;
    currentGOP = gopSize;        /* so first I-frame generates GOP Header */

    /* Rate Control Initialization  */
    bitstreamMode = getRateMode();
    if (bitstreamMode == FIXED_RATE)
      res = initRateControl();

    pastRefFrame = NULL;
    futureRefFrame = NULL;
    for ( i = firstFrame; i <= lastFrame; i++) 
    {
        frameType = FType_Type(i);
	
        /* skip non-reference frames if non-interactive
         * read in non-reference frames if interactive */
        if ( frameType == 'b' ) 
          continue;

        frame = Frame_New(i, frameType);

        pastRefFrame = futureRefFrame;
        futureRefFrame = frame;

	GetNthInputFileName(inputFileName, i);
	if ( childProcess && separateConversion ) 
	  ReadFrame(frame, inputFileName, slaveConversion, TRUE);
	else
	  ReadFrame(frame, inputFileName, inputConversion, TRUE);

        framesRead++;

        if ( ! firstFrameDone ) 
	{
	   char *userData = (char *)NULL;
	   int userDataSize = 0;

	   inputFrameBits = 24*Fsize_x*Fsize_y;
	   SetBlocksPerSlice();
          
          if ( (whichGOP == -1) && (frameStart == -1) ) 
	  {
	     bitstreamMode = getRateMode();
	     if (bitstreamMode == FIXED_RATE) 
	     {
		bit_rate = getBitRate();
		buf_size = getBufferSize();
	     }
	     else 
	     {
		bit_rate = -1;
		buf_size = -1;
	     }
	    
	    if (strlen(userDataFileName) != 0) 
	    {
	       struct stat statbuf;
	       FILE *fp;
	      
	       stat(userDataFileName,&statbuf);
	       userDataSize = statbuf.st_size;
	       userData = malloc(userDataSize);
	       if ((fp = fopen(userDataFileName,"rb")) == NULL) 
	       {
		  fprintf(stderr,"Could not open userdata file-%s.\n",
			userDataFileName);
		  userData = NULL;
		  userDataSize = 0;
		  goto write;
	       }
	       if (fread(userData,1,userDataSize,fp) != userDataSize) 
	       {
		  fprintf(stderr,"Could not read %d bytes from userdata file-%s.\n",
			  userDataSize,userDataFileName);
		  userData = NULL;
		  userDataSize = 0;
		  goto write;
	       }
	    } 
	    else 
	    { 
	       time_t now;
	       time(&now);
	       userData = malloc(100);
	       sprintf(userData,
	         "MPEG stream encoded by UCB Encoder (mpeg_encode) v%s on %s.",
		  VERSION, ctime(&now));
	       userDataSize = strlen(userData);
	    }
          write:
	     Mhead_GenSequenceHeader(bb, Fsize_x, Fsize_y,
                 /* pratio */ aspectRatio,
                 /* pict_rate */ frameRate, /* bit_rate */ bit_rate,
                 /* buf_size */ buf_size, /*c_param_flag */ 1,
                 /* iq_matrix */ qtable, /* niq_matrix */ niqtable,
                 /* ext_data */ NULL, /* ext_data_size */ 0,
                 /* user_data */ userData, /* user_data_size */ userDataSize);
          }
          
	  free(userData);
          firstFrameDone = TRUE;
        }
        
        ProcessRefFrame(frame, bb, lastFrame, outputFileName);

#ifdef GtkPBar
      if(update_PBar((float)(i+1)/(lastFrame+1-firstFrame)) == FALSE)
      {
	 interrupt = TRUE;
	 break;
      }
#endif
     }
    
    if ( frame != NULL )
      Frame_Free(frame);

    /* SEQUENCE END CODE */
    if ( (whichGOP == -1) && (frameStart == -1) )
      Mhead_GenSequenceEnder(bb);

    if ( frameStart == -1 ) 
    {
      /* I think this is right, since (bb == NULL) if (frameStart != -1).
         See above where "bb" is initialized  */
      numBits = bb->cumulativeBits;
    } 
    else {
      /* What should the correct value be?  Most likely 1.  "numBits" is
         used below, so we need to make sure it's properly initialized 
       to somthing (anything).  */
      numBits = 1;
    }

    if ( frameStart == -1 ) 
    {
        Bitio_Flush(bb);
        bb = NULL;
        fclose(ofp);

    } 

/*    if ( (! childProcess) && showBitRatePerFrame )
      CloseBitRateFile();
*/
    if(interrupt)
      return -1;
    else
      return 0;
}



/*===========================================================================*
 *
 * SetStatFileName
 *
 *	set the statistics file name
 *
 * RETURNS:	nothing
 *
 * SIDE EFFECTS:    statFileName
 *
 *===========================================================================*/
void
SetStatFileName(fileName)
    char *fileName;
{
    strcpy(statFileName, fileName);
}


/*===========================================================================*
 *
 * SetGOPSize
 *
 *	set the GOP size (frames per GOP)
 *
 * RETURNS:	nothing
 *
 * SIDE EFFECTS:    gopSize
 *
 *===========================================================================*/
void
SetGOPSize(size)
    int size;
{
    gopSize = size;
}


/*=====================*
 * INTERNAL PROCEDURES *
 *=====================*/

/*===========================================================================*
 *
 * ProcessRefFrame
 *
 *	process an I or P frame -- encode it, and process any B frames that
 *	we can now
 *
 * RETURNS:	nothing
 *
 * SIDE EFFECTS:    stuff appended to bb
 *
 *===========================================================================*/
static void
ProcessRefFrame(frame, bb, lastFrame, outputFileName)
   MpegFrame *frame;
   BitBucket *bb;
   int lastFrame;
   char *outputFileName;
{
   MpegFrame *bFrame = NULL;
   char    fileName[1024];
   char    inputFileName[1024];
   FILE    *fpointer = NULL;
   boolean separateFiles;
   int	    id;
   time_t  tempTimeStart, tempTimeEnd;
  
   separateFiles = (bb == NULL);
   
   if ( separateFiles && (frame->id >= realStart) &&
       (frame->id <= realEnd) ) 
   {
      sprintf(fileName, "%s.frame.%d", outputFileName, frame->id);
      if ( (fpointer = fopen(fileName, "wb")) == NULL ) 
	exit(1);
      
      bb = Bitio_New(fpointer);
   }
   
   /* nothing to do */
   if ( frame->id < realStart )
      return;
   
  /* first, output this frame */
   if ( frame->type == TYPE_IFRAME )
   {
      
      /* only start a new GOP with I */
      /* don't start GOP if only doing frames */
      if ( (! separateFiles) && (currentGOP >= gopSize) ) 
      {
	 int closed;
	 static int num_gop = 0;
	
	 /* first, check to see if closed GOP */
	 if ( totalFramesSent == frame->id || pastRefFrame == NULL)
	    closed = 1;
	 else
	   closed = 0;
	
	 /* new GOP */
	 if (num_gop != 0 && mult_seq_headers && num_gop % mult_seq_headers==0)
	 {
	    Mhead_GenSequenceHeader(bb, Fsize_x, Fsize_y,
		/* pratio */    aspectRatio,
		/* pict_rate */ frameRate, /* bit_rate */ bit_rate,
		/* buf_size */  buf_size,  /* c_param_flag */ 1,
		/* iq_matrix */ customQtable, /* niq_matrix */ customNIQtable,
		/* ext_data */ NULL,  /* ext_data_size */ 0,
		/* user_data */ NULL, /* user_data_size */ 0);
	 }

	
	 num_gop++;
	 Mhead_GenGOPHeader(bb,	/* drop_frame_flag */ 0,
			 tc_hrs, tc_min, tc_sec, tc_pict,
			 closed, /* broken_link */ 0,
			 /* ext_data */ NULL, /* ext_data_size */ 0,
			 /* user_data */ NULL, /* user_data_size */ 0);
	 currentGOP -= gopSize;

      }
      
      if ( (frame->id >= realStart) && (frame->id <= realEnd) ) 
      {
	 GenIFrame(bb, frame);
	
	 framesOutput++;
	
	 if ( separateFiles ) 
	 {
	    Bitio_Flush(bb);
	    fclose(fpointer);
	 }
      }
	 
      numI--;
      timeMask &= 0x6;
	 
      currentGOP++;
   } 
   else 
   {
      if ( (frame->id >= realStart) && (frame->id <= realEnd) ) 
      {
	 GenPFrame(bb, frame, pastRefFrame);
	 
	 framesOutput++;
	 
	 if ( separateFiles ) 
	 {
	    Bitio_Flush(bb);
	    fclose(fpointer);
	 }
      }
      
      numP--;
      timeMask &= 0x5;
      
      currentGOP++;
   }
   
   /* now, output B-frames */
   if ( pastRefFrame != NULL ) 
   {
      for ( id = pastRefFrame->id+1; id < futureRefFrame->id; id++ ) 
      {
	 if ( ! ((id >= realStart) && (id <= realEnd)) )
	   continue;
	 
	 bFrame = Frame_New(id, 'b');
	 
	 /* read B frame, output it */
	 GetNthInputFileName(inputFileName, id);
	 if ( childProcess && separateConversion ) 
	   ReadFrame(bFrame, inputFileName, slaveConversion, TRUE);
	 else 
	   ReadFrame(bFrame, inputFileName, inputConversion, TRUE);

	 framesRead++;
	
	 if ( separateFiles ) 
	 {
	    sprintf(fileName, "%s.frame.%d", outputFileName, 
		    bFrame->id);
	    if ( (fpointer = fopen(fileName, "wb")) == NULL ) 
	      exit(1);

	    bb = Bitio_New(fpointer);
	 }
	 
	 GenBFrame(bb, bFrame, pastRefFrame, futureRefFrame);
	 framesOutput++;
	 
	 if ( separateFiles ) 
	 {
	    Bitio_Flush(bb);
	    fclose(fpointer);
	 }
	 
	 /* free this B frame right away */
	 Frame_Free(bFrame);
	 
	 numB--;
	 timeMask &= 0x3;
	 
	 currentGOP++;
      }
   } 
   else 
   {
      /* SRS replicated code */
      for ( id = 0; id < futureRefFrame->id; id++ ) 
      {
	 if ( ! ((id >= realStart) && (id <= realEnd)) )
	   continue;
	 
	 bFrame = Frame_New(id, 'b');
	    
	 /* read B frame, output it */
	 
	 GetNthInputFileName(inputFileName, id);
	 if ( childProcess && separateConversion ) 
	   ReadFrame(bFrame, inputFileName, slaveConversion, TRUE);
	 else 
	   ReadFrame(bFrame, inputFileName, inputConversion, TRUE);
	 
	 framesRead++;
	 
	 if ( separateFiles ) 
	 {
	    sprintf(fileName, "%s.frame.%d", outputFileName, 
		    bFrame->id);
	    if ( (fpointer = fopen(fileName, "wb")) == NULL ) 
	      exit(1);

	    bb = Bitio_New(fpointer);
	 }
	 
	 GenBFrame(bb, bFrame, (MpegFrame *)NULL, futureRefFrame);
	 framesOutput++;
	 
	 if ( separateFiles ) 
	 {
	    Bitio_Flush(bb);
	    fclose(fpointer);
	 }
	 
	 /* free this B frame right away */
	 Frame_Free(bFrame);
	 
	 numB--;
	 timeMask &= 0x3;
	 
	 currentGOP++;
      }
   }
   
   /* now free previous frame, if there was one */
   if ( pastRefFrame != NULL ) 
     Frame_Free(pastRefFrame);
   
   /* note, we may still not free last frame if lastFrame is incorrect
    * (if the last frames are B frames, they aren't output!)
    */
}


void
ReadDecodedRefFrame(frame, frameNumber)
    MpegFrame *frame;
    int frameNumber;
{
    FILE    *fpointer;
    char    fileName[256];
    int	width, height;
    register int y;

    width = Fsize_x;
    height = Fsize_y;

    sprintf(fileName, "%s.decoded.%d", outputFileName, frameNumber);
    if ((fpointer = fopen(fileName, "rb")) == NULL) 
    {
       sleep(1);
       if ((fpointer = fopen(fileName, "rb")) == NULL)
	 exit(1);
    }

    Frame_AllocDecoded(frame, TRUE);
    
    for ( y = 0; y < height; y++ ) 
    {
      if (fread(frame->decoded_y[y], 1, width, fpointer) != width) 
	fprintf(stderr, "Could not read enough bytes from %s\n", fileName);
    }
    
    for (y = 0; y < (height >> 1); y++) /* U */
    {			
      if (fread(frame->decoded_cb[y], 1, width >> 1, fpointer) != (width>>1)) 
	fprintf(stderr, "Could not read enough bytes from %s\n", fileName);
    }
    
    for (y = 0; y < (height >> 1); y++) /* V */
    {			
      if (fread(frame->decoded_cr[y], 1, width >> 1, fpointer) != (width>>1)) 
	fprintf(stderr, "Could not read enough bytes from %s\n", fileName);
    }
    
    fclose(fpointer);
}

