/*===========================================================================*
 * other_paramet.c                                                           *
 *                                                                           *
 *      Procedures to create or update some parameters Frame                 *
 *                                                                           *
 * EXPORTED PROCEDURES:                                                      *
 *      void createOtherParametFrame(GtkWidget *)                            *
 *      void updateOtherParametFrame()                                       *
 *                                                            (by A. Miyake) *
 *===========================================================================*/
#include <gtk/gtk.h>
#include <math.h>
#include "cono_map.h"
#include "widget_set.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "intl.h"

/*==================*
 * GLOBAL VARIABLES *   
 *==================*/
extern GtkStyle* GreekStyle;
extern GtkTooltips *tooltips;

/*==================*
 * GLOBAL FUNCTIONS *   
 *==================*/
extern void updateStereoFrame();

/*==================*
 * STATIC VARIABLES *
 *==================*/
static GtkWidget *OtherParametFrame, *OtherParametBox;
static GtkWidget *ImageSizeSpin = NULL;

/*====================*
 * Internal Procedure *
 *====================*/
static void setParamet(GtkSpinButton*, gfloat*);
static void setNA(GtkSpinButton*, gfloat*);
static void setImageSize();  /* only for Circular or Animation Mode */


/*=====================*
 * EXPORTED PROCEDURES *
 *=====================*/
void createOtherParametFrame(GtkWidget *parent);
void updateOtherParametFrame();

void  createOtherParametFrame(GtkWidget *parent)
{
   GtkWidget *NASpin, *ThicknessSpin;

   OtherParametFrame = gtk_frame_new(_("(Other Parameters)"));
   gtk_container_set_border_width( GTK_CONTAINER(OtherParametFrame), 10);
   gtk_box_pack_start(GTK_BOX(parent), OtherParametFrame, FALSE, FALSE, 0);

   OtherParametBox  = gtk_vbox_new(FALSE, 0);
   gtk_container_add(GTK_CONTAINER(OtherParametFrame), OtherParametBox);
   
   NASpin = pack_spin(OtherParametBox, _("Numerical Aperture  "), 155, 53, 
		      parameter.NA, 0.1, 1.0, 0.01, 0.1, 2);   
   ThicknessSpin = pack_spin(OtherParametBox, _("Thickness(micron)  "), 155, 
			     53, parameter.thickness, 10, 3000, 1, 10, 0);

   /*********** signnal connects ******************************/
   gtk_signal_connect(GTK_OBJECT (NASpin), "changed",
		      GTK_SIGNAL_FUNC (setNA), &(parameter.NA));
   gtk_signal_connect(GTK_OBJECT (ThicknessSpin), "changed",
		      GTK_SIGNAL_FUNC (setParamet), &(parameter.thickness));

   updateOtherParametFrame();
}

void updateOtherParametFrame()
{
   gint image_size=0, max_image_size=0;
   gfloat* image_size_ptr = NULL;

   if(ImageSizeSpin != NULL)
      gtk_widget_destroy(ImageSizeSpin->parent);

   if(drawing_method.drawing_kind == Single) 
   {
      parameter.image_size = parameter.image_size_for_Single;
      max_image_size = 400;
      image_size_ptr = &(parameter.image_size_for_Single);
   }
   else if (drawing_method.drawing_kind == Circular) 
   {
      parameter.image_size = parameter.image_size_for_Circular;
      image_size = parameter.image_size_for_Circular;
      max_image_size 
	= 660/(1+1/sin(M_PI*parameter.sampling_step_for_Circular/360));
      image_size_ptr = &(parameter.image_size_for_Circular);
   } 
   else if (drawing_method.drawing_kind == Animation)
   {
      parameter.image_size = parameter.image_size_for_Animation;
      max_image_size = 1600/sqrt(360/parameter.sampling_step_for_Animation);
      image_size_ptr = &(parameter.image_size_for_Animation);
   }

   if(parameter.image_size > max_image_size) 
     parameter.image_size = *image_size_ptr = max_image_size;

   ImageSizeSpin = pack_spin(OtherParametBox, _("Image size(pixels)  "), 155,
                             53, parameter.image_size, 20, max_image_size, 
			     1, 10, 0);

   gtk_signal_connect(GTK_OBJECT (ImageSizeSpin), "changed",
		      GTK_SIGNAL_FUNC (setImageSize), image_size_ptr);

   gtk_widget_show_all(OtherParametFrame);
}   

/*===========================================================================*
 * (INTERNAL PROCEDURES)                                                     *
 *                      setParamet                                           *
 *                      setNA                                                *
 *                      setImageSize                                         *
 *===========================================================================*/
static void 
setParamet(GtkSpinButton *spin, gfloat* var)
{
   *var = gtk_spin_button_get_value_as_float(spin);
}

static void 
setNA(GtkSpinButton *spin, gfloat* var)
{
   setParamet(spin, var);
   updateStereoFrame();
}

static void
setImageSize(GtkSpinButton *spin, gfloat* var)
{
   setParamet(spin, var);

   if(drawing_method.drawing_kind == Single) 
     parameter.image_size = parameter.image_size_for_Single;
   else if(drawing_method.drawing_kind == Circular) 
     parameter.image_size = parameter.image_size_for_Circular;
   if(drawing_method.drawing_kind == Animation) 
     parameter.image_size = parameter.image_size_for_Animation;
}
